Module error_minimization

    !Find the parameters of best-fit for the core-shell particle
    
    use working_precision
    use RI_functions
    use shell_modes
    
Implicit None

real(wp), parameter, private :: Pi = 3.14159265358979323846_wp !pi
    
Contains
 
Subroutine Find_pars(rpar,k,v_exp,v_calc,cModes,r,S,r_steps,guess_s,guess_c,K_s,K_c,ModeSet,y)

    real(wp), parameter :: DEL = 1.0E-2_wp !The difference used when calculating p and q
    integer, parameter :: CUT_OFF = 8 !maximum number of loops
    real(wp), parameter :: TOL = 1.0E-2_wp !The tolerance for S to change

    real(wp), parameter :: nm = 1.0_wp !the RI of the medium

    real(wp), allocatable :: v_exp(:) !A list that contains the wavenumbers of the current frame
    real(wp), allocatable :: v_calc(:,:) !The calculated wavenumbers

    real(wp), allocatable :: r(:) !contains r
    real(wp), allocatable :: S(:) !S for all r
    integer :: r_steps !number of r steps
    integer, allocatable :: ModeSet(:,:) !the current mode set being fit
    real(wp), allocatable :: y(:,:) !the linewidths in size parameter

    real(wp), allocatable :: rpar(:,:) !contains the solved parameters for each r
    							!(r,outer radius, mc0, mc1, ... , ms0, ms1, ...)  

    integer :: k !The total number of parameters that are being solved for
    
    real(wp) :: rpar_best(k) !stores the parameters of best-fit
    real(wp) :: v_calc_best(cModes) !best wavenumbers
    real(wp) :: y_best(cModes) !best linewidths

    integer :: cModes !Number of modes in current frame

    real(wp) :: guess_s, guess_c !first guesses for the RI of core and shell
    !The RI is a Cauchy expression: m0 + m1*v^2 + m2*v^4 + ... + mK*v^(2K)
    integer :: K_s, K_c !K for shell and core
    real(wp) :: m_c(cModes), m_s(cModes) !The RI at each v for the core and shell
    real(wp) :: a_calc !the calculated radius

  
    real(wp) :: S_best !use in the loop when calculating parameters 
    complex(wp) :: p(cModes), q(cModes) ! partial derivatives
    real(wp) :: b(cModes) ! linear intercept !p, q, and b used in the error minimization
    real(wp) :: d(k), A(k,k) !the vector d and matrix A
  
    real(wp) :: p_plus, p_minus !used to speedup the calculation of one resonance
  
    complex(wp) :: z !temporarily stores the complex resonance position 
  	    
    integer :: N,L,T !mode number, order, and polarization
 
    character, parameter :: cr = achar(13) !carriage return 
    real(wp) :: per_com !percent complete
  
    real(wp), parameter :: RI_LB = 1.25_wp, RI_UB = 10.0_wp !lower and upper bounds on possible RI  
    integer :: i,j,w,v
  
  
    !set the solution vector to 0
    rpar = 0.0_wp
  
  	!use the initial guesses for the RI as initial values for the solution vector
    Do i = 1, r_steps
        rpar(i,1) = 1.0_wp ! convenient
        rpar(i,2) = guess_c
        rpar(i,3+K_c) = guess_s
    End do
	
    S = huge(1.0_wp) 

	!begin loop
    Do i = 1, r_steps
  		
  		!these three lines just output the progress of the calculation
        per_com = 100*(r(i) - r(1))/(r(r_steps)-r(1))
        write(*, 200, advance='NO') r(i), r(r_steps), per_com,cr
        200 format(TL10,"Fitting ratio ", F8.6," of ",F8.6,", ",F5.1," percent complete"  A)
  		
        j=0

        S_best = huge(1.0_wp)

        Do while (j.LT.CUT_OFF) !loop until j gets too big
        
        If (j.NE.0) Then ! do not compute in first loop
        
            Call Calc_Ad(A,d,k,REAL(p),REAL(q),b,v_exp,cModes,K_s,K_c)
            
            Call LU_decomposition(A,d,rpar(i,:),k)
            
        end if
        
    		 
			!Find m_c and m_s with input of par and v_exp
            do w = 1, cModes
                m_c(w) = Cauchy_RI(K_c,rpar(i,2:2+K_c),v_exp(w))
                m_s(w) = Cauchy_RI(K_s,rpar(i,3+K_c:3+K_c+K_s),v_exp(w))
				!write(*,*) v_exp(w),m_s(w)
                If (m_c(w).LT.RI_LB.OR.m_c(w).GT.RI_UB) Then ! if the values found are out of bounds, we re-initialize them
                    rpar(i,2) = guess_c
                    rpar(i,3:2+K_c) = 0.0_wp
                    do v =1, w
                        m_c(v) = Cauchy_RI(K_c,rpar(i,2:2+K_c),v_exp(v))
                    end do
                end if
                If (m_s(w).LT.RI_LB.OR.m_s(w).GT.RI_UB) Then
                    rpar(i,3+K_c) = guess_s
                    rpar(i,4+K_c:3+K_c+K_s) = 0.0_wp
                    do v =1, w
                        m_s(v) = Cauchy_RI(K_s,rpar(i,3+K_c:3+K_c+K_s),v_exp(v))
                    end do
                end if
            end do
            
			!Calculate p, q, and b with current RI
            do w = 1, cModes
				
                N = ModeSet(w,1)
                L = ModeSet(w,2)
                T = ModeSet(w,3)
				
                !z = ShRes(N,L,T,r(i),m_c(w),m_s(w),nm)
                
                Call ShRes(z,p(w),q(w),N,L,T,r(i),m_c(w),m_s(w),nm)
                
                a_calc = rpar(i,1)
                
                v_calc(i,w) = real(z)/(2*Pi*a_calc)
                
                y(i,w) = aimag(z)
                
                !CALL ResEqDerRI(p(w),q(w),N,T,z,r(i),m_c(w),m_s(w),nm)
                
                b(w) = REAL(z - p(w)*m_c(w) - q(w)*m_s(w))
                
                end do
                
            If (j.EQ.0) then
                !just in case the very first run through gives S(i) as NaN
                rpar_best = rpar(i,:)
                v_calc_best = v_calc(i,:)
                y_best = y(i,:)
            End if
        
        	!Calculate S
           S(i) = 0.0_wp
           do w = 1, cModes
                S(i) = S(i) + (v_exp(w) - v_calc(i,w))**2   
            end do
        	
           If (S_best.GT.S(i)) then
                S_best = S(i)
                rpar_best = rpar(i,:)
                v_calc_best = v_calc(i,:)
                y_best = y(i,: )
                j = j +1
                Else
                j = CUT_OFF	
                End if
            End do
	    
	    !calculate the wavenumbers for the new parameters 
		!do w = 1, cModes
		!	v_calc(i,w) = v_calc_best(w)
	    !   	y(i,w) = y_best(w)
		!end do
		
        S(i) = S_best
	      
    End do


End Subroutine Find_pars

Subroutine Verify_Calc(A,d,k,p,q,b,v_exp,cModes,par,rt)

	!Only called during testing

  	integer :: k !The total number of parameters that are being solved for
	integer :: cModes !Number of modes in current frame
	
  	real(wp) :: d(k), A(k,k) !the vector d and matrix A
  	real(wp) :: p(cModes), q(cModes), b(cModes) !p, q, and b used in the error minimization

	real(wp) :: par(k) !the parameters

    real(wp), allocatable :: v_exp(:) !A list that contains the wavenumbers of the current frame

	character :: tab = ACHAR(9) !tab

	integer :: rt

	integer :: n,m,j

	If (rt.EQ.0) Then
	Open (7, FILE = "bqpv.txt", action="Write",status="replace")
    do j = 1, cModes
    	Write(7,*) b(j),tab,q(j),tab,p(j),tab,v_exp(j)
    end do
    Close(7)

	Open (7, FILE = "d.txt", action="Write",status="replace")
    do j = 1, k
    	Write(7,*) d(j)
    end do
    Close(7)

	Open (7, FILE = "A.txt", action="Write",status="replace")
    do m = 1, k
    	Write(7,*) (A(m,n),tab,n=1,k)
    end do
    Close(7)

	Else

	Open (7, FILE = "pars.txt", action="Write",status="replace")
    do j = 1, k
    	Write(7,*) par(j)
    end do
    Close(7)
	

	read(*,*)
    End if 

End Subroutine Verify_Calc

Subroutine Calc_Ad(A,d,k,p,q,b,v_exp,cModes,K_s,K_c)

  	integer :: k !The total number of parameters that are being solved for
	integer :: cModes !Number of modes in current frame
	
  	real(wp) :: d(k), A(k,k) !the vector d and matrix A
  	real(wp) :: p(cModes), q(cModes), b(cModes) !p, q, and b used in the error minimization

  	integer :: K_s, K_c !K for shell and core

    real(wp), allocatable :: v_exp(:) !A list that contains the wavenumbers of the current frame

	integer :: n,m,j,w,u

	A = 0.0_wp
	d = 0.0_wp
	
	
	!Find d
	do j = 1, cModes
		d(1) = d(1) + b(j)**2
	end do	

	do n = 2, 2+K_c
		w = n-2
		do j = 1, cModes
			d(n) = d(n) + b(j)*p(j)*(v_exp(j)**(2*w))
		end do	
	end do

	do n = 3+K_c, 3+K_c+K_s
		w = n-(3+K_c)
		do j = 1, cModes
			d(n) = d(n) + b(j)*q(j)*(v_exp(j)**(2*w))
		end do	
	end do

	!make all d's negative
	d = -d

	!Find A
	!A11
	do j = 1, cModes
		A(1,1) = A(1,1) + v_exp(j)*b(j)
	end do	
	A(1,1) = -2*Pi*A(1,1)

	!A(1,core)
	do n = 2, 2+K_c
		w = n-2
		do j = 1, cModes
			A(1,n) = A(1,n) + b(j)*p(j)*(v_exp(j)**(2*w))
		end do	
	end do
	
	!A(1,shell)
	do n = 3+K_c, 3+K_c+K_s
		w = n-(3+K_c)
		do j = 1, cModes
			A(1,n) = A(1,n) + b(j)*q(j)*(v_exp(j)**(2*w))
		end do	
	end do

	!A(core,1)
	do n = 2, 2+K_c
		w = n-2
		do j = 1, cModes
			A(n,1) = A(n,1) + p(j)*(v_exp(j)**(2*w+1))
		end do
		A(n,1) = -2*Pi*A(n,1)
	end do

	!A(shell,1)
	do n = 3+K_c, 3+K_c+K_s
		w = n-(3+K_c)
		do j = 1, cModes
			A(n,1) = A(n,1) + q(j)*(v_exp(j)**(2*w+1))
		end do
		A(n,1) = -2*Pi*A(n,1)
	end do
	
	!A(core,core)
	do n = 2, 2+K_c
		w = n-2
		do m = 2, 2+K_c
			u = m-2
			do j = 1, cModes
				A(n,m) = A(n,m) + (p(j)**2)*(v_exp(j)**(2*w+2*u))
			end do
		end do 
	end do

	!A(shell,shell)
	do n = 3+K_c, 3+K_c+K_s
		w = n-(3+K_c)
		do m = 3+K_c, 3+K_c+K_s
			u = m-(3+K_c)
			do j = 1, cModes
				A(n,m) = A(n,m) + (q(j)**2)*(v_exp(j)**(2*w+2*u))
			end do
		end do 
	end do

	!A(core,shell) and (shell,core)
	do n = 2, 2+K_c
		w = n-2
		do m = 3+K_c, 3+K_c+K_s
			u = m-(3+K_c)
			do j = 1, cModes
				A(n,m) = A(n,m) + q(j)*p(j)*(v_exp(j)**(2*w+2*u))
			end do
			A(m,n) = A(n,m)
		end do 
	end do



End Subroutine Calc_Ad

Subroutine LU_decomposition(A,d,p,k)
    
    !Solve A.p=d using LU decomposition followed by forward and backward substitution
    !Solved as follows:
    !A.p=d, trying to solve p
    !when A=LU then d=L.y
    !where y=U.p
    !so find y using L & d then p using U & y
    !Note: the code is written in such a way that A gets altered
    
    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + m2*(v-v0)^k)
    real(wp) :: A(k,k), d(k) !the matrix A and vector d
	real(wp) :: p(k) !parameters of best-fit
    real(wp) :: L(k,k), U(k,k) !the matrices L and U for the decomposition
  	real(wp) :: y(k) !y vector used during forward substitution 
    real(wp) :: c
    integer :: i,j,m,n
    
    n=k
    
    L=0.0_wp
    U=0.0_wp
    y=0.0_wp
    p=0.0_wp
    
    !Find the lower, off-diagonal elements for L
    do m=1,n-1
        do i=m+1,n
            c = A(i,m)/A(m,m)
            L(i,m) = c
            do j=m+1,n
                A(i,j) = A(i,j) - c*A(m,j)
            end do
        end do
    end do
    
    !Add the identity matrix
    do i=1,n
        L(i,i) = 1.0_wp
    end do
    
	!The upper matrix U
	do j=1,n
        do i=1,j
            U(i,j) = A(i,j)
        end do
    end do
    
    !Forward substitution
    !Find y    
    y(1)=d(1)/L(1,1)
    do i=2,n
    	do j=1,i-1
			y(i) = y(i) - L(i,j)*y(j)
		end do
		y(i) = (y(i) + d(i))/L(i,i)    	
    end do
    
    !Backward substitution
    !Find p
    p(n)=y(n)/U(n,n)
    do i=n-1,1,-1
		do j=n,i+1,-1
    		p(i) = p(i) - U(i,j)*p(j)
    	end do
    	p(i) = (p(i) + y(i))/U(i,i)
    end do
    
End subroutine

End Module error_minimization