Program Main

    !Program: MRSFIT (Mie Resonance Shell Fitting)
    !Authors: Benjamin Vennes & Thomas C. Preston 
    !Contact: thomas.preston@mcgill.ca

 	!Version 1.0: January.31.2019 
 
    !Description:
    !Finds the size and RI of a core-shell particle sphere by fitting Mie resonances. 
    !The RI is defined using a Cauchy type equation
    !m = m0 + m1*v^2 + m2*v^4 + ... + mk*v^(2k) 
    
    !For more details see: 
    !README.txt
    
    !All code written by BV and TCP is covered by the GNU General Public License. 
    !The only code not written by BV and TCP is the complex Bessel function module, 
    !complexbessel.f90, see module header for source.
    use working_precision
    use read_and_write
    use shell_modes
    use error_minimization
    Implicit None
  
    character :: peaklistname*100 !name of the file with all the observed modes in it.
    character :: outputname*100 !name of the output file
    character :: altoutputname*115 !name of the output file with index and radius only
    
    real(wp), allocatable :: medRI(:) !The RI of the medium
    real(wp), allocatable :: obsModes(:) !A list that contains ALL the observed modes (in wavenumber)
    real(wp), allocatable :: v_exp(:) !A list that contains the wavenumbers of the current frame
    real(wp), allocatable :: v_calc(:,:) !The calculated wavenumbers
	real(wp), allocatable :: y(:,:) !the linewidths in size parameter

    real(wp), allocatable :: rpar(:,:) !contains the solved parameters for each r
    							!(r,outer radius, mc0, mc1, ... , ms0, ms1, ...)	
    real(wp), allocatable :: r(:) !contains r
    real(wp), allocatable :: S(:) !S for all r
    integer, allocatable :: nModes(:,:) !the number of modes per frame and the starting posn in the obsModes list
    integer, allocatable :: ModeSet(:,:) !the current mode set being fit
    integer, allocatable :: Known(:,:) !If the mode assignment is known it can be used as an input parameter, this stores it.
    
    !real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.


    !The RI is a Cauchy expression: m0 + m1*v^2 + m2*v^4 + ... + mK*v^(2K)
  	integer :: K_s, K_c !K for shell and core

	integer :: k !The total number of parameters that are being solved for

	real(wp) :: guess_s, guess_c !first guesses for the RI of core and shell

	real(wp) :: r_min, r_max, r_i !ratio: min, max, step size
	integer :: r_steps !number of r steps

    real(wp) :: t1, t2 !used to track time
    real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.
    integer :: cModes, sPosn !Number of modes in current frame and the starting position in obsModes
    integer :: nFrames !the total number of frames
    integer :: sframe, fframe, iframe, cframe !starting frame, ending frame, interval for frames
    
    integer :: i, j    
    complex(wp) :: V1, P1 ,Q1 ! used for testing
    !Read input files  
    Call ReadInput(peaklistname,outputname,altoutputname,sframe,fframe,iframe, &
                     K_s,K_c,guess_s,guess_c,r_min,r_max,r_i)

    !Import frame list and format
    Call FormatInput(peaklistname,obsModes,nModes,nFrames,Known)
	
 
 	!Initialize the output files
 	Call WriteHeaders(outputname,altoutputname,K_s,K_c)
 
    !Number of frames
    If (fframe.NE.-1) Then
        nframes = fframe
    End if
 
 	!number of steps in r_i
 	r_steps = NInt((r_max - r_min)/r_i) + 1
    
    k = K_s + K_c + 3 
    
    allocate(rpar(r_steps,k))
	allocate(S(r_steps))
	allocate(r(r_steps))
	
	do j = 1, r_steps
		r(j) = r_min + (j-1)*r_i
	end do

    Do cFrame = sframe, nframes, iframe
    
        cModes = nModes(1,cframe) 
        !You need at least k = K_s + K_c + 4 modes in a frame to fit
        If (cModes.GE.k) Then
        	    
            Call cpu_time(t1)
            
            sPosn = nModes(2,cFrame)
            
            allocate(v_exp(cModes))
            allocate(v_calc(r_steps,cModes))
            allocate(y(r_steps,cModes))
            allocate(ModeSet(cModes,3))

            !read the modes of the current frame into the array
            Do i = 1,cModes
                   j = i + sPosn - 1
                v_exp(i) = obsModes(j)
            End do
            
            !Currently we assume that mode assignment is known, so:
            Do i = 1,cModes
                j = i + sPosn - 1
                ModeSet(i,:) = Known(j,:)   
            End do

            !Gives you parameters + a table of S
            Call Find_pars(rpar,k,v_exp,v_calc,cModes,r,S,r_steps,guess_s,guess_c,K_s,K_c,ModeSet,y)

            Call WriteOutput(outputname,altoutputname,cFrame,cModes,k,&
                             ModeSet,v_exp,v_calc,rpar,S,r,r_steps,K_s,K_c,y)
             
            Call WriteSvsr(outputname,cFrame,S,r,r_steps,cModes) 
             
            Call cpu_time(t2)

            Call WriteStatus(cFrame,nFrames,t2-t1)

            deallocate(ModeSet)
            deallocate(v_exp)
            deallocate(v_calc)
            deallocate(y)
        	    
        Else
        
            Call WriteStatus(cFrame,nFrames,-1.0_wp)
        
        End if
    
    End do    
    deallocate(rpar)
    deallocate(S)
    deallocate(r)
    deallocate(nModes)
    deallocate(Known)

    


End Program Main