Module read_and_write

!This module has the subroutines for reading the input and setting up the arrays for calculations

    use working_precision
    use RI_functions
 
Implicit None

real(wp), parameter, private :: Pi = 3.14159265358979323846_wp !pi

Contains

Subroutine ReadInput(peaklistname,outputname,altoutputname,sframe,fframe,iframe, &
                     K_s,K_c,guess_s,guess_c,r_min,r_max,r_i)
                      
    character :: peaklistname*100 !name of the file with all the observed modes in it.
    character :: outputname*100 !name of the output file
    character :: altoutputname*115 !name of the output file with index and radius only
    character :: dummy

    integer :: sframe, fframe, iframe !starting frame, ending frame, interval for frames
    
    !The RI is a Cauchy expression: m0 + m1*v^2 + m2*v^4 + ... + mK*v^(2K)
  	integer :: K_s, K_c !K for shell and core

	real(wp) :: guess_s, guess_c !first guesses for the RI of core and shell

	real(wp) :: r_min, r_max, r_i !ratio: min, max, step size

	!Stores text description of RI functions
	character :: RIcoreText*200
	character :: RIshellText*200

	integer :: j


	Write(*,'(a)') "Running MrsFit"
	Write(*,'(a)') "Reading parameters.in..."

    !Read the input parameters -- parameters.in file

    Open(2, file = 'parameters.in')
    read(2, FMT='(A1)') dummy
    read(2, FMT='(A1)') dummy
    read(2,*) peaklistname
    read(2,*) outputname
    read(2, FMT='(A1)') dummy
    read(2,*) sframe, fframe, iframe
    read(2, FMT='(A1)') dummy
    read(2, FMT='(A1)') dummy
    read(2,*) K_s
    read(2,*) K_c
    read(2, FMT='(A1)') dummy
    read(2,*) guess_s
    read(2,*) guess_c
    read(2, FMT='(A1)') dummy
    read(2,*) r_min, r_max, r_i  
    close(2)

	Write(*,'(a)') ""

	Write(*,'(a)') "Note that the formatting of "//Trim(peaklistname)//" should be:"
	Write(*,'(a)') "[Frame number] [Resonance wavelength] [Mode number] [Mode order] [Polarization]"
	Write(*,'(a)') "Use 0 for TE modes and 1 for TM modes in the Polarization column"

	Write(*,'(a)') ""

    altoutputname = Trim(outputname)//"-IndexAndRadius.txt"
    outputname = Trim(outputname)//".txt"    

	Call RIfunctionText(K_c,K_s,RIcoreText,RIshellText)

    Write(*,'(a)') Trim(RIcoreText)
    Write(*,'(a)') Trim(RIshellText)
    
	Write(*,'(a)') ""
    
    Return
    
End Subroutine ReadInput

Subroutine RIfunctionText(K_c,K_s,RIcoreText,RIshellText)

  	integer :: K_s, K_c !K for shell and core

	!Stores text description of RI functions
	character :: RIcoreText*200
	character :: RIshellText*200

    If (K_s.EQ.0) Then
    	RIshellText = "RIShell = m0s"
    Else if (K_s.EQ.1) Then
    	RIshellText = "RIShell = m0s + m1s*v^2"
	Else if (K_s.EQ.2) Then
		RIshellText = "RIShell = m0s + m1s*v^2 + m2s*v^4"
	Else if (K_s.EQ.3) Then
		RIshellText = "RIShell = m0s + m1s*v^2 + m2s*v^4 + m3s*v^6"
    Else if (K_s.EQ.4) Then
		RIshellText = "RIShell = m0s + m1s*v^2 + m2s*v^4 + m3s*v^6 + m4s*v^8"
	Else if (K_s.EQ.5) Then
		RIshellText = "RIShell = m0s + m1s*v^2 + m2s*v^4 + m3s*v^6 + m4s*v^8 + m5s*v^10"	
    End if    
    
    If (K_c.EQ.0) Then
    	RIcoreText = "RICore = m0c"
    Else if (K_c.EQ.1) Then
    	RIcoreText = "RICore = m0c + m1c*v^2"
	Else if (K_c.EQ.2) Then
		RIcoreText = "RICore = m0c + m1c*v^2 + m2c*v^4"
	Else if (K_c.EQ.3) Then
		RIcoreText = "RICore = m0c + m1c*v^2 + m2c*v^4 + m3c*v^6"
	Else if (K_c.EQ.4) Then
		RIcoreText = "RICore = m0c + m1c*v^2 + m2c*v^4 + m3c*v^6 + m4c*v^8"
	Else if (K_c.EQ.5) Then
		RIcoreText = "RICore = m0c + m1c*v^2 + m2c*v^4 + m3c*v^6 + m4c*v^8 + m5c*v^10"
    End if
    
    Return

End Subroutine RIfunctionText

Subroutine FormatInput(peaklistname,obsModes,nModes,nFrames,Known)

    !Import modes from data file and compare them to calculated modes
    !Also converts modes from wavelength to wavenumber

    character :: peaklistname*100 !name of the file with all the observed modes in it.
    character :: dummy
    
    real(wp), allocatable :: rawDat(:,:) !The raw data from the peak list file
    real(wp), allocatable :: obsModes(:) !A list that contains the observed modes
    integer, allocatable :: nModes(:,:) !the number of modes per frame and the starting posn in the obsModes list
    integer, allocatable :: Known(:,:) !If the mode assignment is known it can be used as an input parameter, this stores it.

    integer :: nMax    !the total number of modes listed in the input file
                    !(the total number of lines in the input file with data minus one)
    integer :: nFrames !the total number of frames
    
    integer :: io, i, j, cFrame
        
    !Count the number of lines in the peak list file.
    nMax=1

    Open(2, file = peaklistname)
	
    Do
        read(2,*,IOSTAT=io) dummy
        If (io < 0) then
            Exit
        Else
            nMax=nMax+1
        End if
    End do
	
    close(2)

    !The data file is such that the last line is blank. Adjust nMax accordingly:
    nMax=nMax-1

    !read the data file into an array
    allocate(rawDat(2,nMax))
	allocate(Known(nMax,3))
	
    Open(2, file = peaklistname)
    
    Do i = 1, nMax
        read(2,*) rawDat(1,i), rawDat(2,i), Known(i,1), Known(i,2), Known(i,3)
        rawDat(2,i) = 1.0_wp/rawDat(2,i) !change to wavenumber     
    End do
	
    close(2)

    !Store the modes in one list
    allocate(obsModes(nMax))

    Do i = 1, nMax
        obsModes(i) = rawDat(2,i)     
    End do

    !Determine the number modes for each frame
    nFrames = Int(rawDat(1,nMax))
    allocate(nModes(2,nFrames))
  
    Do i = 1, nFrames
        nModes(1,i)=0
        nModes(2,i)=1
    End do

    i=1
    cFrame=1

    Do while (i.LE.nMax)
        If(Int(rawDat(1,i)).EQ.cFrame) then
            nModes(1,cFrame)=nModes(1,cFrame)+1
        Else
            cFrame=cFrame+1
            i=i-1
        End if
        i=i+1
    End do

    !Determine where the modes for a frame i start in obsModes
    !For frames with no modes set the starting position to -1

    If (nModes(1,1).EQ.0) Then
        nModes(2,1) = -1
    End if

    Do i = 2, nFrames        
        If (nModes(1,i).EQ.0) Then
            nModes(2,i) = -1
        Else
            Do j = 1, i-1
                nModes(2,i) = nModes(1,j) + nModes(2,i)
            End do
        End if
    End do

    Return
    
End Subroutine FormatInput


Subroutine WriteHeaders(outputname,altoutputname,K_s,K_c)

    character :: outputname*100 !name of the output file
    character :: altoutputname*115 !name of the output file for index and radius only
    character :: tab = ACHAR(9) !tab
    character :: kslist(11)
    character :: kclist(11)
  	integer :: K_s, K_c !K for shell and core
  	
	!Stores text description of RI functions
	character :: RIcoreText*200
	character :: RIshellText*200

    integer :: i
    
    Do i = 0,K_c
        Write(kclist(i+1), '(i1)' ) i
    End do
    
    Do i = 0,K_s
        Write(kslist(i+1), '(i1)' ) i
    End do
    
    
    Open (7, FILE = altoutputname, action="Write",status="replace")
    Write(7,*) "Frame", tab, "Outer Radius", tab, "Inner Radius", tab, "r",tab, &
    		   ("mc",kclist(i),tab, i=1,K_c+1),("ms",kslist(i),tab, i=1,K_s+1), &
    		    "S wavenumber^2",tab,"S wavelength^2",tab,&
    		    "S wavenumber^2/J",tab,"S wavelength^2/J"
    Close(7)

	Call RIfunctionText(K_c,K_s,RIcoreText,RIshellText)
    
    Open (7, FILE = outputname, action="Write",status="replace")
    Write(7,*) "Note: to change the RI function into one of wavelength replace v with 1/wavelength."

    Write(7,*)

    Write(7,*) Trim(RIcoreText)
    Write(7,*) Trim(RIshellText)

    Close(7)

    Return
    
End Subroutine WriteHeaders

Subroutine WriteStatus(cFrame,nFrames,t)

	!Updates the console with the current progress of the program

    character (len=30) :: aOutput !used to format the some console output
    character (len=30) :: bOutput !used to format the some console output
    character (len=30) :: cOutput !used to format the some console output


	real(wp) :: t
	integer :: cFrame, nFrames

	If (t.GT.0.0_wp) Then
    ! Segmentation fault around here
    !Write(*,*) "Hello!"
		Write(aOutput,*) cFrame
		Write(bOutput,*) nFrames
		Write(cOutput,26) t
		26 format (F20.3)
		
		Write(*,'(a)') "Frame: " // trim(adjustl(aOutput)) // " of " // trim(adjustl(bOutput)) //  &
				   " was completed in " // trim(adjustl(cOutput)) // " seconds.              "
	Else
		Write(aOutput,*) cFrame
		Write(bOutput,*) nFrames
		Write(*,'(a)') "Frame: " // trim(adjustl(aOutput)) // " of " // trim(adjustl(bOutput)) // &
				   " had too few modes for the chosen RI function."
    End if
                       
End Subroutine WriteStatus

Subroutine WriteOutput(outputname,altoutputname,cFrame,cModes,k,&
                       ModeSet,v_exp,v_calc,rpar,S,r,r_steps,K_s,K_c,y)

	character :: outputname*100 !name of the output file
	character :: altoutputname*115 !name of the output file for modes
    character :: tab = ACHAR(9) !tab
	character :: mType*2 !polarization TE or TM
	
	integer :: cFrame !current frame number
	integer :: cModes !Number of modes in current frame and the starting position in obsModes

    real(wp), allocatable :: v_exp(:) !A list that contains the wavenumbers of the current frame
   	real(wp), allocatable :: v_calc(:,:) !The calculated wavenumbers
    real(wp), allocatable :: rpar(:,:) !contains the solved parameters for each r
    							!(r,outer radius, mc0, mc1, ... , ms0, ms1, ...) 
	real(wp), allocatable :: y(:,:) !the linewidths in size parameter
  
    real(wp), allocatable :: r(:) !contains r
    real(wp), allocatable :: S(:) !S for all r
  	integer :: r_steps !number of r steps
    integer, allocatable :: ModeSet(:,:) !the current mode set being fit
  
  	integer :: k !The total number of parameters that are being solved for

  	integer :: K_s, K_c !K for shell and core

	character (len=30) :: f(13) !used to format the output (eliminate white spaces)
	character (len=30) :: par_out(k) !used to format the output (eliminate white spaces)

	integer :: min_posn !the location of the minimum value of S
	
	real(wp) :: S_v, S_wav !S in wavevnumber and wavelength
	
	real(wp) :: vEr, wEr !store difference in wavenumber and wavelength
	real(wp) :: line_v, line_w !current linewidths
	real(wp) :: v,x,vm,vp !used when calculating linewidths
	real(wp) :: curr_RIc, curr_RIs !the RI of the core and shell at the current wavenumber
	
	integer :: i,j,N,L,T

	min_posn = minloc(S,1)

    Do i = 1, k
    	Write(par_out(i),*) rpar(min_posn,i)
    End do
	
	S_v = 0.0_wp 
	S_wav  = 0.0_wp
	
	Do i = 1, cModes
		S_v = S_v + (v_exp(i) - v_calc(min_posn,i))**2
		S_wav = S_wav + (1.0_wp/v_exp(i) - 1.0_wp/v_calc(min_posn,i))**2
	End do

	Write(f(1),*) cFrame
	Write(f(2),*) rpar(min_posn,1)*r(min_posn)
	Write(f(3),*) r(min_posn)
	Write(f(4),*) S_v
	Write(f(5),*) S_wav
	Write(f(6),*) S_v/cModes
	Write(f(7),*) S_wav/cModes

    Open (8, FILE = altoutputname, access="append") 
    Write(8,*) trim(adjustl(f(1))),tab,trim(adjustl(par_out((1)))),tab,&
    		   trim(adjustl(f(2))),tab,trim(adjustl(f(3))),tab,&
    		   (trim(adjustl(par_out((i)))),tab,i=2,k), &
    		   trim(adjustl(f(4))),tab,trim(adjustl(f(5))),tab, &
    		   trim(adjustl(f(6))),tab,trim(adjustl(f(7)))
    Close(8)

	Open (8, FILE = outputname, access="append")
	Write(8,*)
	Write(f(1),*) cFrame
	Write(8,*) "Frame:", tab, trim(adjustl(f(1)))
	Write(8,*) "Outer radius, inner radius, and ratio:", tab, trim(adjustl(par_out((1)))), tab, &
				trim(adjustl(f(2))), tab, trim(adjustl(f(3)))
	Write(8,*) "Core parameters:", (tab,trim(adjustl(par_out((i)))),i=2,2+K_c)
	Write(8,*) "Shell parameters:", (tab,trim(adjustl(par_out((i)))),i=3+K_c,3+K_c+K_s)
	Write(8,*)
    Write(8,*) "Mode assignment"
    Write(8,*) "Mode number",tab,"Mode order",tab,"Polarization",tab,"Wavenumber (Obs.)",tab,"Wavenumber (Calc.)",tab, &
               "Error",tab,"Wavelength (Obs.)",tab,"Wavelength (Calc.)",tab,"Error",tab,"Core RI",tab,"Shell RI", tab, &
               "Linewidth (wavenumber)",tab,"Linewidth (wavelength)" 
	
	Do i = 1,cModes
        N=ModeSet(i,1)
        L=ModeSet(i,2)
        T=ModeSet(i,3)
        
        If (T.EQ.0) Then
            mType = "TE"
        Else
            mType = "TM"
        End if
	
		vEr = Abs(v_exp(i)-v_calc(min_posn,i))
		wEr = Abs(1.0_wp/v_exp(i)-1.0_wp/v_calc(min_posn,i))
	
		curr_RIc = Cauchy_RI(K_c,rpar(min_posn,2:2+K_c),v_exp(i))
		curr_RIs = Cauchy_RI(K_s,rpar(min_posn,3+K_c:3+K_c+K_s),v_exp(i))
	
		x = 2*Pi*rpar(min_posn,1)*v_calc(min_posn,i)
	    vm=(x+y(min_posn,i))/(2*Pi*rpar(min_posn,1))
        vp=(x-y(min_posn,i))/(2*Pi*rpar(min_posn,1))
	
		line_v = vp-vm
		line_w = (1.0_wp/vm - 1.0_wp/vp)
	
		!Format text for output
		Write(f(1),*) N
		Write(f(2),*) L
		Write(f(3),*) mType
		Write(f(4),*) v_exp(i)
		Write(f(5),*) v_calc(min_posn,i)
		Write(f(6),*) vEr
		Write(f(7),*) 1.0_wp/v_exp(i)
		Write(f(8),*) 1.0_wp/v_calc(min_posn,i)
		Write(f(9),*) wEr
		Write(f(10),*) curr_RIc
		Write(f(11),*) curr_RIs
		Write(f(12),*) line_v
		Write(f(13),*) line_w
		
		Write(8,*) (trim(adjustl(f(j))),tab,j=1,13)
	End do
	
	Write(f(1),*) S_v
	Write(f(2),*) S_wav
	Write(f(3),*) S_v/cModes
	Write(f(4),*) S_wav/cModes
	Write(8,*)
    Write(8,*) "S -- Wavenumber^2:",tab,trim(adjustl(f(1))),tab,"Wavelength^2:",tab,trim(adjustl(f(2)))
	Write(8,*) "S/J -- Wavenumber^2:",tab,trim(adjustl(f(3))),tab,"Wavelength^2:",tab,trim(adjustl(f(4)))
	
	Close(8)

End Subroutine WriteOutput

Subroutine WriteSvsr(outputname,cFrame,S,r,r_steps,cModes)

	!Outputs the S as function of r. Used mainly for testing.

	character :: outputname*100 !name of the output file
	character :: Svsr_outputname*115 !name of the output file for modes
    character :: tab = ACHAR(9) !tab

	integer :: cModes !Number of modes in current frame and the starting position in obsModes

	integer :: cFrame !current frame number


  
    real(wp), allocatable :: r(:) !contains r
    real(wp), allocatable :: S(:) !S for all r
  	integer :: r_steps !number of r steps
 
	character (len=30) :: f(13) !used to format the output (eliminate white spaces)

	integer :: i

	Write(f(1),*) cFrame


	Svsr_outputname = "Svsr-"//trim(adjustl(f(1)))//"-"//outputname

    Open (7, FILE = Svsr_outputname, action="Write",status="replace")
    
    Write(7,*) "r", tab, "S/J Wavenumber^2"
    Do i = 1, r_steps
		If (S(i).LT.huge(1.0_wp)) Then
	    	Write(7,*) r(i),tab,S(i)/cModes
		End if
	end do
     
    Close(7)

End Subroutine WriteSvsr

! 
! Subroutine WriteOutput(outputname,altoutputname,cFrame,cModes,p,k,ModeSet,vexp,Airy,v0,fCau,medRI,ks)
!     
!     character :: outputname*100 !name of the output file
!     character :: altoutputname*115 !name of the output file for modes
!     character :: tab = ACHAR(9) !tab
!     character :: mType*2 !polarization TE or TM
!     
!     character (len=30) :: f(13) !used to format the output (eliminate white spaces)
!     character (len=30) :: pout(k+2) 
!     
!     real(wp), allocatable :: medRI(:) !The RI of the medium
!     real(wp), allocatable :: vexp(:) !A list that contains the wavenumbers of the current frame
!     real(wp) :: v0,v0t !The midpoint of the spectrum -- wavenumber
!     integer, allocatable :: ModeSet(:,:) !the overall best mode set
!     real(wp) :: p(k+2) !contains the solved parameters (radius, m0, m1, ...)
!     real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.
!     complex(wp) :: z !complex size parameter
!     real(wp) :: x,y !real and imaginary part of size parameter
!     real(wp) :: v !x/(2*Pi*a)
!     real(wp) :: lwv, lww !linewidth in wavenumber and wavelength
!     real(wp) :: vm,vp !used when calculating the linewidth
!     real(wp) :: vEr,wEr,tvEr,twEr !errors in the wavenumber and wavelength and their totals
!     real(wp) :: cRI,relRI !current RI and relative RI of the sphere (they're the same when the RI of the medium is 1)
!     integer :: cFrame !current frame number
!     integer :: cModes !Number of modes in current frame and the starting position in obsModes
!     integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + m2*(v-v0)^k)
!     integer :: ks  !Number of terms in the refractive index of medium (ms0 + ms1*v + ... + msk*v^ks)
!     integer :: fCau !Use Cauchy RI formula (0:No, 1:Yes)
!     
!     integer :: i,j,N,L,T
!         
!     Call midPoint(v0,v0t,vexp,cModes)
!     
!     tvEr=0.0_wp
!     twEr=0.0_wp
!     
!     Open (8, FILE = outputname, access="append")  
!         
!     Write(f(1),*) cFrame
!     Do i=1,k+2
!     	Write(pout(i),*) p(i)
!     End do
!     Write(8,*) "Frame:", tab, trim(adjustl(f(1)))
!     Write(8,*) "Parameters (radius, m0, m1, ...)"
!     Write(8,*) (trim(adjustl(pout(i))),tab,i=1,k+2)
!     Write(8,*)
!     Write(f(1),*) v0t
!     Write(f(2),*) 1.0_wp/v0t   
!     Write(8,*) "Midpoint -- Wavenumber:",tab,trim(adjustl(f(1))),tab,"Wavelength:",tab,trim(adjustl(f(2)))
!     Write(8,*)
!     Write(8,*) "Mode assignment"
!     Write(8,*) "Mode number",tab,"Mode order",tab,"Polarization",tab,"Wavenumber (Obs.)",tab,"Wavenumber (Calc.)",tab, &
!                "Error",tab,"Wavelength (Obs.)",tab,"Wavelength (Calc.)",tab,"Error",tab,"RI",tab,"Relative RI", tab, &
!                "Linewidth (wavenumber)",tab,"Linewidth (wavelength)" 
!     
!     Do i = 1,cModes
!         N=ModeSet(i,1)
!         L=ModeSet(i,2)
!         T=ModeSet(i,3)
!         
!         If (T.EQ.0) Then
!             mType = "TE"
!         Else
!             mType = "TM"
!         End if
!         
!         !vexp(i)=vexp(i)*RIs(medRI,ks,vexp(i))
!         relRI=fRI(p,k,vexp(i),v0t,fCau)
!         z=MieRes(N,Airy(L),relRI,T)
!         x=real(z)
!         y=aimag(z)
! 
!         v=x/(2*Pi*p(1))
!         
!         vm=(x+y)/(2*Pi*p(1))
!         vp=(x-y)/(2*Pi*p(1))
!         
!         !adjust for RI
!         cRI=relRI*RIs(medRI,ks,vexp(i))
!         v=v/RIs(medRI,ks,vexp(i))
!         vexp(i)=vexp(i)/RIs(medRI,ks,vexp(i))
!         
!         vEr=Abs(vexp(i)-v)
!         wEr=Abs(1/vexp(i)-1/v)
!         
!         lwv = (vp-vm)/RIs(medRI,ks,vexp(i))
!         lww = RIs(medRI,ks,vexp(i))*(1.0_wp/vm - 1.0_wp/vp) !CHECK
!         
!         !Format text for output
!         Write(f(1),*) N
!         Write(f(2),*) L
!         Write(f(3),*) mType
!         Write(f(4),*) vexp(i)
!         Write(f(5),*) v
!         Write(f(6),*) vEr
!         Write(f(7),*) 1.0_wp/vexp(i)
!         Write(f(8),*) 1.0_wp/v
!         Write(f(9),*) wEr
!         Write(f(10),*) cRI
! 		Write(f(11),*) relRI
! 		Write(f(12),*) lwv
! 		Write(f(13),*) lww
! 
! 		Write(8,*) (trim(adjustl(f(j))),tab,j=1,13)
!         !Write(8,*) N,tab,L,tab,mType,tab,vexp(i),tab,v,tab,vEr,tab,1.0_wp/vexp(i),tab,1.0_wp/v,tab,wEr,tab,cRI,tab,relRI
!         tvEr=tvEr+vEr**2
!         twEr=twEr+wEr**2
!     End do
!     
!     tvEr=tvEr/cModes
!     twEr=twEr/cModes
!     
!     Write(f(1),*) tvEr
!     Write(f(2),*) twEr
!     
!     Write(8,*)
!     Write(8,*) "Mean error squared -- Wavenumber^2:",tab,trim(adjustl(f(1))),tab,"Wavelength^2:",tab,trim(adjustl(f(2)))
!     Write(8,*)
!     
!     Close(8)
! 
! 	!Second output file
!     Write(f(1),*) cFrame
!     Write(f(2),*) v0t
!     Write(f(3),*) 1.0_wp/v0t
!     Write(f(4),*) tvEr
!     Write(f(5),*) twEr
!     
!     Open (8, FILE = altoutputname, access="append") 
!     Write(8,*) trim(adjustl(f(1))),tab,(trim(adjustl(pout(i))),tab,i=1,k+2),(trim(adjustl(f(i))),tab,i=2,5)
!     
!     Close(8)
!     
! End subroutine

End Module read_and_write