Module shell_modes
    
    !find the resonances of a shell using Equation 2 from: 
	!Moridnejad, A., Preston, T. C., & Krieger, U. K. (2017). J. Phys. Chem. A, 121(42), 8176–8184. 
    !http://doi.org/10.1021/acs.jpca.7b06774
    !find the resonances of a shell using Equation 7 from: 
    ! Vennes, B., Preston T. C., (2019) JOSA(A)
	!this resonance condition is found using the denominators of
    !Equations 34 and 35 in Aden & Kerker J. App. Phys. 22, 1951, 1242
    !the implementation here only allows for real refractive indices

    use working_precision
    use complexbessel
    use sphere_modes

Implicit none

real(wp), parameter, private :: Pi = 3.14159265358979323846_wp !pi

Contains

subroutine ShRes(v1,p1,q1,n,l,pType,f0,m1,m2,m3)

	!returns the complex shell resonance
	
    integer :: n !the mode number
    integer :: l !the mode order
    integer :: pType !type of polarization (0: a/TE, 1: b/TM -- polarization definitions are different from modern convention)
	
    complex(wp) :: v1, p1, q1

    real(wp) :: m1, m2, m3 !RI of the core, shell and medium

    real(wp) :: f0 !the ratio of alpha/nu at resonance (core/shell)

	!If a faster method is developed it can be replaced here
    !ShRes = ShResRI(n,l,pType,f0,m1,m2,m3)
    
    Call ShResRI(v1,p1,q1,n,l,ptype,f0,m1,m2,m3) ! now a subroutine because it outputs size parameter and derivatives

End subroutine

subroutine ShResRI(v1,p1,q1,n,l,pType,f0,m1,m2,m3)

    complex(wp) :: v1, p1, q1 ! size parameter and partial derivatives
    
    complex(wp) :: v0, v2

    integer :: n, l, pType ! mode assignments are the mode number, order, and polarization type (TE == 0 TM == 1)

    real(wp) :: f0, m1, m2, m3 ! core/shell ratio ... RI of the core, shell, and medium

    integer :: steps ! number of steps to reach the final point

    real(wp) :: dm ! incremental change in shell RI

    real(wp) :: m ! the current shell RI
    
    real(wp) :: del1, del2 ! check on the difference
    
    real(wp), parameter :: THRESH_DEL1 = 1.0_wp, THRESH_DEL2 = 0.0_wp ! THRESH_DEL1 compares the absolute value of the difference to unity.
                                                                      ! ThRESH_DEL2 compares the sign of the change to the linear projection 
    
    complex(wp) :: q0! analytic derivative of z wrt m2
    
    real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.
    
    integer :: j
    
    Call AiryRoots(Airy) ! needed for MieRes
    
    ! increment and steps
    
    dm = 0.01_wp
    steps = INT((m2-m1)/(m3*dm))
    steps = Abs(steps)
    If ((m2-m1).LT.0.0_wp) Then
        dm = -dm
    End if
    
    !initial guess
    
    v0 = MieRes(n,Airy(l),m1/m3,pType) ! homogeneous size parameter with core refractive index
    v1 = v0
    v2 = v0
    q0 = 0.0_wp
    m = m1
    do j = 1, steps
        m = m1 + dm*j ! shell RI increment
        v1 = v0 ! update
        CALL NR_Iteration_1D(n,pType,v1,p1,q1,f0,m1,m,m3,1) ! solver
        del1 = Abs((v1-v0)) ! Check for errors
        del2 = REAL(q0)*dm*REAL(v1-v2)
        If (del1.GT.THRESH_DEL1.OR.del2.LT.THRESH_DEL2) Then
            v1 = ShResRIBackItup(n,pType,v2,q0,f0,m1,m,m3,dm) ! backup recursion
            CALL ResEqDerRI(p1,q1,n,pType,v1,f0,m1,m,m3) ! derivative
        end if
        v0 = v1 + dm*q1 ! linear projection forward
        v2 = v1
        q0 = q1
    end do
    
    ! solve the last step
    dm = (m2 - m)
    v0 = v1 + dm*q0
    v1 = v0
    CALL NR_Iteration_1D(n,pType,v1,p1,q1,f0,m1,m2,m3,2)
    
    !del1 = Abs(REAL(v1-v0))
    !If (del1.GT.THRESH_DEL1) Then
    !        v1 = ShResRIBackItup(n,pType,v2,f0,m1,m2,m3,dm)
    !end if
    
End subroutine
    
Recursive function ShResRIBackItup(n,pType,v0,q0,f0,m1,m2,m3,dm) result(v1)

    ! this is a recursive function in case the iteration fails for the given increment in refractive index
    ! THRESH defines the smallest increment desired, if the stepsize is smaller, the recursion function gives up
    ! steps is the number of desired steps to divide the increment into.

    complex(wp) :: v0, v1, v2

    integer :: n, pType ! mode assignments are the mode number, order, and polarization type (TE == 0 TM == 1)

    real(wp) :: f0, m1, m2, m3 ! core/shell ratio ... RI of the core, shell, and medium

    integer, parameter :: steps = 10 ! number of steps to reach the final point

    real(wp) :: dm, dma ! incremental change in shell RI

    real(wp) :: m, mi ! the current shell RI
    
    real(wp) :: del1, del2 ! check on the difference
    
    real(wp), parameter :: THRESH_DEL1 = 1.0_wp, THRESH_DEL2 = 0.0_wp, THRESH = 1.0E-4_wp
    
    complex(wp) :: p, q0, q1 ! analytic derivative of z wrt m2
    
    integer :: j
    
    mi = m2 - dm ! back it up
    dma = dm/(steps) ! divide it by (step) bits
    
    !Call ResEqDerRI(p,q0,n,pType,v0,f0,m1,mi,m3)
    v2 = v0
    v0 = v0 + q0*dma
    m = mi
    do j = 1, steps
        m = mi + dma*j
        v1 = v0
        call NR_Iteration_1D(n,pType,v1,p,q1,f0,m1,m,m3,1) ! find the root
        del1 = Abs((v1-v0))
        del2 = REAL(q0)*dma*REAL(v1-v2)
        If (del1.GT.THRESH_DEL1.OR.del2.LT.THRESH_DEL2) Then ! check on the residual
            If (Abs(dm).GT.THRESH) Then
                v1 = ShResRIBackItup(n,pType,v2,q0,f0,m1,m,m3,dma)
            else
                v1 = v0    
            end if  
            CALL ResEqDerRI(p,q1,n,pType,v1,f0,m1,m,m3) ! linear projection
        end if
        v2 = v1
        v0 = v1 + dma*q1
        q0 = q1
    end do
    
    CALL NR_Iteration_1D(n,pType,v1,p,q1,f0,m1,m2,m3,0) ! solve last step
        
End function
    
    
Subroutine NR_Iteration_1D(n,pType,v1,p1,q1,f0,m1,m2,m3,par) ! CHECKED

    ! Here, we find the root of the resonant equation using the 1D Newton Method
    !Newton Raphson in 1-d is: z1 = z0 - ratio where ratio=g/(f0*ga + gv)
    
    integer :: n, pType, par ! mode number and polarization TE TM, par is discussed in ResEq
    
    complex(wp) :: v0, v1, p1, q1, ratio ! output resonance and derivatives
    
    real(wp) :: f0 ! core/shell ratio
    
    real(wp) :: m1, m2, m3 ! RI of the core, shell, and medium
        
    integer :: iter ! current of number of iterations
    
    integer, parameter :: maxiter = 1000 ! iteration number and maximum threshold
    
    real(wp), parameter :: maxtol = 1.0E-8_wp ! maximum tolerance
    
    iter = 1 ! initialize
    
    v0 = huge(1.0_wp) 
    
    Do while (Abs(v1-v0).GT.maxtol.AND.iter.LT.maxiter) ! loop over    
        v0 = v1
        Call ResEq(ratio,p1,q1,n,pType,v0,f0,m1,m2,m3,0) ! get ratio only
        v1 = v0 - ratio
        iter = iter + 1    
    End do
    Call ResEq(ratio,p1,q1,n,pType,v1,f0,m1,m2,m3,par) ! get derivative depending on par value
    
end subroutine
    
subroutine ResEq(ratio,p,q,n,pType,v1,f0,m1,m2,m3,par) ! CHECKED

    ! This subroutine evaluates the resonant equation
    ! as well as the partial derivatives of the size parameter
    ! using the formalism in Vennes/Preston. 
    ! ratio is used in the 1D Newton's method
    ! p and q are partial derivatives (dz/dmc, dz/dms) 
    ! none are calculated when par == 0, q is calculated when par == 1
    ! and both are calculated when par == 2
    ! default is zero
    
    complex(wp) :: g, ga, gv, gb, gw
    
    complex(wp) :: ratio, p, q ! g/(f0*ga + gv), and partial derivatives (dz/dmc = -gb/(f0*ga + gv), dz/dms = -gw/(f0*ga + gv)) respectively
    
    real(wp) :: m1, m2, m3 ! RI of the core, shell and medium

    real(wp) :: f0 !the ratio of alpha/nu at resonance

    complex(wp) :: a1, v1 !the initial guess
    
    integer :: n,pType,par !the mode number
    
    complex(wp) :: zc, zs ! core and shell complex size parameter
    
    real(wp) :: mc,ms ! core and shell real refractive index
    
    real(wp) :: v, w ! TE (1,1) or TM [(1/ms)^2, (mc/ms)^2]
    
    complex(wp) :: mczc, mszs, mszc ! to reduce cluttering
    
    complex(wp) :: ric_mszc, ric_mszs, ricJsc, ricYsc, ricJss, ricYss ! to reduce cluttering

    complex(wp) :: Aw, Bv ! to reduce cluttering
    
    complex(wp) :: Psi_cc, Psi_ss, Psi_sc, Chi_ss, Chi_sc, Xi_s ! log. derivatives
    
    complex(wp) :: Jcc(-1:0), Jsc(-1:0), Ysc(-1:0), Jss(-1:0), Yss(-1:0), Hs(-1:0) ! subroutine calls
    
    complex(wp) :: bTmp(2) !temporary stores the bessel function result
    integer :: order, error, nout !parameters used for the Bessel function module
    real(wp) :: bOrd
    
    integer :: i

    bOrd = real(n-1,wp)+0.5_wp ! What I need is to compute J,Y, and H at n-1+0.5, and n + 0.5
    order = 2 
    
    mc = m1/m3 ! relative refractive indices
    ms = m2/m3
    
    If (pType.EQ.0) Then ! TE polarization
        v = 1.0_wp
        w = 1.0_wp
    else                 ! TM polarization
        v = (1/ms)**2
        w = (mc/ms)**2
    End if   
    
    a1=v1*f0 !initial guess for a1 is just v1 x the ratio of the core to shell
    
    zc = a1 ! size parameters
    zs = v1

    mczc = mc*zc ! convenient to have
    mszc = ms*zc
    mszs = ms*zs

    ric_mszc = zsqrt(pi*mszc/2.0_wp) ! coefficient in front of the ricatti bessel function
    ric_mszs = zsqrt(pi*mszs/2.0_wp)

    !Testing to make sure Bessel functions give right values
    !mczc = (2.0_wp,0.1_wp)
    !Call cbesh(mczc, bOrd, 1, 2, order, bTmp, nout, error)
    !write(*,*) Sqrt(Pi/2.0_wp)*zsqrt(1.0_wp/(mczc))*bTmp(1) 
    !write(*,*) Sqrt(Pi/2.0_wp)*zsqrt(1.0_wp/(mczc))*bTmp(2) 
    
    !Build all the Bessel functions    
    Call cbesj(mczc, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2 
        Jcc(i-2) = bTmp(i) ! no need for the constant in front ... We do not need it for log. der and we have them already for the ric-bess functions
    End do
    Call cbesj(mszc, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Jsc(i-2) = bTmp(i)
    End do
    Call cbesy(mszc, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Ysc(i-2) = bTmp(i)
    End do
    Call cbesj(mszs, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Jss(i-2) = bTmp(i)
    End do
    Call cbesy(mszs, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Yss(i-2) = bTmp(i)
    End do
    Call cbesh(zs, bOrd, 1, 1, order, bTmp, nout, error)
    Do i = 1,2
        Hs(i-2) = bTmp(i)
    End do
    
	! Construct the log. derivatives
    ! the subscript xx indicates the argument of the log. derivatives (m_x z_x)

    Psi_cc = (mc/ms)*(Jcc(-1)/Jcc(0) - n/mczc)
    Psi_sc = Jsc(-1)/Jsc(0) - n/mszc
    Chi_sc = Ysc(-1)/Ysc(0) - n/mszc
    Psi_ss = Jss(-1)/Jss(0) - n/mszs
    Chi_ss = Yss(-1)/Yss(0) - n/mszs
    Xi_s = (Hs(-1)/Hs(0) - n/zs)/ms
        
    ! Construct the RB functions
        
    ricJsc = ric_mszc*Jsc(0)
    ricJss = ric_mszs*Jss(0)
    ricYss = ric_mszs*Yss(0)
    ricYsc = ric_mszc*Ysc(0)
        
    Aw = (w-(mc/ms)**2) + (1.0_wp-w)*n*(n+1)/(mszc**2) ! convenient coefficients to have
    Bv = (v-(1.0_wp/ms)**2) + (1.0_wp-v)*n*(n+1)/(mszs**2)
    g = ( (ricJsc*(Psi_cc - w*Psi_sc)) * (ricYss*(Xi_s - v*Chi_ss)) & ! see Appendix B
        - (ricYsc*(Psi_cc - w*Chi_sc)) * (ricJss*(Xi_s - v*Psi_ss)) )
    ga = ms*(ricJsc*(Psi_cc*(Psi_sc - Psi_cc) + Aw ) ) &
            *(ricYss*(Xi_s - v*Chi_ss)) - ms*(ricJss*(Xi_s - v*Psi_ss)) &
            *(ricYsc*(Psi_cc* (Chi_sc - Psi_cc) + Aw ))
    gv = ms*(ricYss*(Xi_s* (Chi_ss - Xi_s) + Bv )) &
            *(ricJsc*(Psi_cc - w*Psi_sc)) - ms*(ricYsc*(Psi_cc - w*Chi_sc)) &
            *(ricJss*(Xi_s* (Psi_ss - Xi_s) + Bv ))
    
    ratio = g/(f0*ga + gv)
    
    ! Here, we get derivatives for free
    
    If (par.EQ.0) THEN ! compute none
        gb = 0.0_wp
        gw = 0.0_wp
        p = 0.0_wp
        q = 0.0_wp
    End If
    If (par.EQ.1) THEN ! compute q
        gb = 0.0_wp
        If (pType.EQ.0) Then
            gw = ricJsc*ricYss & ! Appendix C
                * ((Psi_cc*(mszc*Psi_sc - 1.0_wp) + w*( mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Chi_ss) + (Psi_cc - w*Psi_sc)*( Xi_s*(mszs*Chi_ss - 1.0_wp) &
                + v*( mszs - n*(n+1)/mszs)) )/ms - ricJss*ricYsc &
                * ( (Psi_cc*(mszc*Chi_sc - 1.0_wp) + w*( mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Psi_ss) + (Psi_cc - w*Chi_sc)*( Xi_s*(mszs*Psi_ss - 1.0_wp) &
                + v*( mszs - n*(n+1)/mszs)) )/ms
        else 
            gw = ricJsc*ricYss &
                *( (Psi_cc*(mszc*Psi_sc - 1.0_wp) + w*(2.0_wp*Psi_sc +  mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Chi_ss) + (Psi_cc - w*Psi_sc)*( Xi_s*(mszs*Chi_ss - 1.0_wp) &
                + v*(2.0_wp*Chi_ss + mszs - n*(n+1)/mszs)) )/ms - ricJss*ricYsc &
                *( (Psi_cc*(mszc*Chi_sc - 1.0_wp) + w*(2.0_wp*Chi_sc + mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Psi_ss) + (Psi_cc - w*Chi_sc)*( Xi_s*(mszs*Psi_ss - 1.0_wp) &
                + v*(2.0_wp*Psi_ss + mszs - n*(n+1)/mszs)) )/ms;
        End if
            p = 0.0_wp
            q = - gw/(f0*ga + gv)
    End if
    If (par.EQ.2) THEN! compute p and q
        If (pType.EQ.0) Then ! Appendix C
            gb = ricJsc*(Psi_cc*(1.0_wp-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc )/mc &
                *ricYss*(Xi_s - v*Chi_ss) - ricJss*(Xi_s - v*Psi_ss) &
                *ricYsc*(Psi_cc*(1-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc )/mc
    
            gw = ricJsc*ricYss &
                * ((Psi_cc*(mszc*Psi_sc - 1.0_wp) + w*( mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Chi_ss) + (Psi_cc - w*Psi_sc)*( Xi_s*(mszs*Chi_ss - 1.0_wp) &
                + v*( mszs - n*(n+1)/mszs)) )/ms - ricJss*ricYsc &
                * ( (Psi_cc*(mszc*Chi_sc - 1.0_wp) + w*( mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Psi_ss) + (Psi_cc - w*Chi_sc)*( Xi_s*(mszs*Psi_ss - 1.0_wp) &
                + v*( mszs - n*(n+1)/mszs)) )/ms
            else
                gb = ricJsc*(Psi_cc*(1.0_wp-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc - 2.0_wp*w*Psi_sc)/mc &
                    *ricYss*(Xi_s - v*Chi_ss) - ricJss*(Xi_s - v*Psi_ss) &
                    *ricYsc*(Psi_cc*(1.0_wp-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc - 2.0_wp*w*Chi_sc)/mc
    
                gw = ricJsc*ricYss &
                    *( (Psi_cc*(mszc*Psi_sc - 1.0_wp) + w*(2.0_wp*Psi_sc +  mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Chi_ss) + (Psi_cc - w*Psi_sc)*( Xi_s*(mszs*Chi_ss - 1.0_wp) &
                + v*(2.0_wp*Chi_ss + mszs - n*(n+1)/mszs)) )/ms - ricJss*ricYsc &
                *( (Psi_cc*(mszc*Chi_sc - 1.0_wp) + w*(2.0_wp*Chi_sc + mszc - n*(n+1)/mszc)) &
                * (Xi_s - v*Psi_ss) + (Psi_cc - w*Chi_sc)*( Xi_s*(mszs*Psi_ss - 1.0_wp) &
                + v*(2.0_wp*Psi_ss + mszs - n*(n+1)/mszs)) )/ms;
            End if
    
            p = - gb/(f0*ga + gv) ! chain rule: Appendix C
            q = - gw/(f0*ga + gv)
    End If
    
    
End subroutine

Subroutine ResEqDerRI(p,q,n,pType,v1,f0,m1,m2,m3) ! This subroutine is only needed in case of failure

    complex(wp) :: p,q ! output dzs_dmc dzs_dms
    
    complex(wp) :: ga, gb, gv, gw ! output g, ga, gv

    real(wp) :: m1, m2, m3 ! RI of the core, shell and medium

    real(wp) :: f0 !the ratio of alpha/nu at resonance

	complex(wp) :: a1, v1 !the initial guess

    integer :: pType,n !the mode number
    
	complex(wp) :: zc, zs ! core/shell complex size parameter

	real(wp) :: mc,ms ! core and shell real refractive index

	real(wp) :: v, w ! TE (1,1) or TM [(1/ms)^2, (mc/ms)^2]

	complex(wp) :: mczc, mszs, mszc, ric_mszc, ric_mszs, ricJsc, ricYsc, ricJss, ricYss ! to reduce cluttering

	complex(wp) :: Aw, Bv ! to reduce cluttering

	complex(wp) :: Psi_cc, Psi_ss, Psi_sc, Chi_ss, Chi_sc, Xi_s ! log. derivatives

	complex(wp) :: Jcc(-1:0), Jsc(-1:0), Ysc(-1:0), Jss(-1:0), Yss(-1:0), Hs(-1:0) ! subroutine calls

	complex(wp) :: bTmp(2) !temporary stores the bessel function result
    integer :: order, error, nout !parameters used for the Bessel function module
    real(wp) :: bOrd
    integer :: i
    
    !Parameters used for 2D Newton's method
    !a1 and v1 are provided as initial guesses for alpha and nu
    !nu, the size parameter k3*b, b is the outer radius
    !alpha, the size parameter k3*a, a is the inner radius
    
    bOrd = real(n-1,wp)+0.5_wp ! What I need is to compute J,Y, and H at n-1+0.5, and n + 0.5
    order = 2 
    
	mc = m1/m3
    ms = m2/m3
    
    If (pType.EQ.0) Then
        v = 1
        w = 1
    else
        v = (1/ms)**2
        w = (mc/ms)**2
    End if   
    
    a1=v1*f0 !initial guess for a1 is just v1 x the ratio of the core to shell
    
    zc = a1
	zs = v1

    mczc = mc*zc ! convenient to have
    mszc = ms*zc
    mszs = ms*zs

    ric_mszc = zsqrt(pi*mszc/2.0_wp)
    ric_mszs = zsqrt(pi*mszs/2.0_wp)

    !Build all the Bessel functions    
    Call cbesj(mczc, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2 
        Jcc(i-2) = bTmp(i) ! no need for the constant in front ... We do not need it for log. der and we have them already for the ric-bess functions
    End do
    Call cbesj(mszc, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Jsc(i-2) = bTmp(i)
    End do
    Call cbesy(mszc, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Ysc(i-2) = bTmp(i)
    End do
    Call cbesj(mszs, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Jss(i-2) = bTmp(i)
    End do
    Call cbesy(mszs, bOrd, 1, order, bTmp, nout, error)
    Do i = 1,2
        Yss(i-2) = bTmp(i)
    End do
    Call cbesh(zs, bOrd, 1, 1, order, bTmp, nout, error)
    Do i = 1,2
        Hs(i-2) = bTmp(i)
    End do
		! Construct the log. derivatives

    Psi_cc = (mc/ms)*(Jcc(-1)/Jcc(0) - n/mczc)
    Psi_sc = Jsc(-1)/Jsc(0) - n/mszc
    Chi_sc = Ysc(-1)/Ysc(0) - n/mszc
    Psi_ss = Jss(-1)/Jss(0) - n/mszs
    Chi_ss = Yss(-1)/Yss(0) - n/mszs
    Xi_s = (Hs(-1)/Hs(0) - n/zs)/ms
        
    ! Construct the RB functions
        
    ricJsc = ric_mszc*Jsc(0)
    ricJss = ric_mszs*Jss(0)
    ricYss = ric_mszs*Yss(0)
    ricYsc = ric_mszc*Ysc(0)
        
    Aw = (w-(mc/ms)**2) + (1.0_wp-w)*n*(n+1)/(mszc**2)
    Bv = (v-(1.0_wp/ms)**2) + (1.0_wp-v)*n*(n+1)/(mszs**2)
    ga = ms*(ricJsc*(Psi_cc*(Psi_sc - Psi_cc) + Aw ) ) &
            *(ricYss*(Xi_s - v*Chi_ss)) - ms*(ricJss*(Xi_s - v*Psi_ss)) &
            *(ricYsc*(Psi_cc* (Chi_sc - Psi_cc) + Aw ))
    gv = ms*(ricYss*(Xi_s* (Chi_ss - Xi_s) + Bv )) &
            *(ricJsc*(Psi_cc - w*Psi_sc)) - ms*(ricYsc*(Psi_cc - w*Chi_sc)) &
            *(ricJss*(Xi_s* (Psi_ss - Xi_s) + Bv ))
            
    If (pType.EQ.0) Then
    gb = ricJsc*(Psi_cc*(1.0_wp-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc )/mc &
        *ricYss*(Xi_s - v*Chi_ss) - ricJss*(Xi_s - v*Psi_ss) &
        *ricYsc*(Psi_cc*(1-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc )/mc
    
    gw = ricJsc*ricYss &
        * ((Psi_cc*(mszc*Psi_sc - 1.0_wp) + w*( mszc - n*(n+1)/mszc)) &
        * (Xi_s - v*Chi_ss) + (Psi_cc - w*Psi_sc)*( Xi_s*(mszs*Chi_ss - 1.0_wp) &
        + v*( mszs - n*(n+1)/mszs)) )/ms - ricJss*ricYsc &
        * ( (Psi_cc*(mszc*Chi_sc - 1.0_wp) + w*( mszc - n*(n+1)/mszc)) &
        * (Xi_s - v*Psi_ss) + (Psi_cc - w*Chi_sc)*( Xi_s*(mszs*Psi_ss - 1.0_wp) &
        + v*( mszs - n*(n+1)/mszs)) )/ms
    else
    gb = ricJsc*(Psi_cc*(1.0_wp-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc - 2.0_wp*w*Psi_sc)/mc &
        *ricYss*(Xi_s - v*Chi_ss) - ricJss*(Xi_s - v*Psi_ss) &
        *ricYsc*(Psi_cc*(1.0_wp-mszc*Psi_cc) + n*(n+1)/mszc - mc/ms*mczc - 2.0_wp*w*Chi_sc)/mc
    
    gw = ricJsc*ricYss &
        *( (Psi_cc*(mszc*Psi_sc - 1.0_wp) + w*(2.0_wp*Psi_sc +  mszc - n*(n+1)/mszc)) &
        * (Xi_s - v*Chi_ss) + (Psi_cc - w*Psi_sc)*( Xi_s*(mszs*Chi_ss - 1.0_wp) &
        + v*(2.0_wp*Chi_ss + mszs - n*(n+1)/mszs)) )/ms - ricJss*ricYsc &
        *( (Psi_cc*(mszc*Chi_sc - 1.0_wp) + w*(2.0_wp*Chi_sc + mszc - n*(n+1)/mszc)) &
        * (Xi_s - v*Psi_ss) + (Psi_cc - w*Chi_sc)*( Xi_s*(mszs*Psi_ss - 1.0_wp) &
        + v*(2.0_wp*Psi_ss + mszs - n*(n+1)/mszs)) )/ms;
    End if
    
    p = - gb/(f0*ga + gv) ! chain rule
    q = - gw/(f0*ga + gv)
    
End subroutine


End module